"use strict";

var modules = require("./ui.grid_core.modules"),
    utils = require("../filter_builder/utils"),
    gridCoreUtils = require("./ui.grid_core.utils"),
    customOperations = require("./ui.grid_core.filter_custom_operations");

var FILTER_ROW_OPERATIONS = ["=", "<>", "<", "<=", ">", ">=", "notcontains", "contains", "startswith", "endswith", "between"],
    HEADER_FILTER_OPERATIONS = ["anyof", "noneof"];

var FilterSyncController = modules.Controller.inherit(function () {
    var setHeaderFilterValue = function setHeaderFilterValue(columnsController, column, headerFilterCondition) {
        if (headerFilterCondition) {
            columnsController.columnOption(column.dataField, {
                filterType: headerFilterCondition[1] === "anyof" ? "include" : "exclude",
                filterValues: headerFilterCondition[2]
            });
        }
    };

    var clearHeaderFilterValues = function clearHeaderFilterValues(columnsController, column) {
        columnsController.columnOption(column.dataField, {
            filterType: "include",
            filterValues: undefined
        });
    };

    var setFilterRowCondition = function setFilterRowCondition(columnsController, column, filterValue, selectedFilterOperation) {
        columnsController.columnOption(column.dataField, {
            filterValue: filterValue,
            selectedFilterOperation: selectedFilterOperation
        });
    };

    var clearFilterRowCondition = function clearFilterRowCondition(columnsController, column) {
        columnsController.columnOption(column.dataField, {
            filterValue: undefined,
            selectedFilterOperation: undefined
        });
    };

    return {
        syncFilterValue: function syncFilterValue(isInit) {
            var that = this,
                columnsController = that.getController("columns"),
                columns = columnsController.getColumns(),
                filterValue = that.option("filterValue");

            columns.forEach(function (column) {
                var headerFilterCondition = utils.getMatchedCondition(filterValue, column.dataField, HEADER_FILTER_OPERATIONS);
                if (headerFilterCondition) {
                    setHeaderFilterValue(columnsController, column, headerFilterCondition);
                } else if (column.filterValues && !isInit) {
                    clearHeaderFilterValues(columnsController, column);
                }

                var filterRowCondition = utils.getMatchedCondition(filterValue, column.dataField, FILTER_ROW_OPERATIONS);
                if (filterRowCondition) {
                    setFilterRowCondition(columnsController, column, filterRowCondition[2], filterRowCondition[1]);
                } else if (column.filterValue && !isInit) {
                    clearFilterRowCondition(columnsController, column);
                }

                if (column.filterValues && !headerFilterCondition) {
                    filterValue = that._getSyncHeaderFilter(filterValue, column);
                }

                if (column.filterValue && !filterRowCondition) {
                    filterValue = that._getSyncFilterRow(filterValue, column, column.filterValue);
                }
            });

            if (filterValue && filterValue.length > 0) {
                that._setFilterValue(filterValue);
            }
        },

        init: function init() {
            if (this.getController("data").skipCalculateColumnFilters()) {
                this.syncFilterValue(true);
            }
        },

        _setFilterValue: function _setFilterValue(filterValue) {
            this._skipSyncFilterValue = true;
            this.option("filterValue", filterValue);
            this._skipSyncFilterValue = false;
        },

        _getSyncFilterRow: function _getSyncFilterRow(filterValue, column, value) {
            var operation = column.selectedFilterOperation || column.defaultFilterOperation || utils.getDefaultOperation(column),
                filter = [column.dataField, operation, value || null];
            return utils.syncFilters(filterValue, filter, FILTER_ROW_OPERATIONS);
        },

        _getSyncHeaderFilter: function _getSyncHeaderFilter(filterValue, column) {
            var filterValues = column.filterValues,
                filter = [column.dataField];
            if (filterValues) {
                filter.push(column.filterType === "exclude" ? "noneof" : "anyof");
                filter.push(filterValues);
            } else {
                filter.push("=");
                filter.push(null);
            }
            return utils.syncFilters(filterValue, filter, HEADER_FILTER_OPERATIONS);
        },

        syncFilterRow: function syncFilterRow(column, value) {
            this._setFilterValue(this._getSyncFilterRow(this.option("filterValue"), column, value));
        },

        syncHeaderFilter: function syncHeaderFilter(column) {
            this._setFilterValue(this._getSyncHeaderFilter(this.option("filterValue"), column));
        },

        getCustomFilterOperations: function getCustomFilterOperations() {
            var filterBuilderCustomOperations = this.option("filterBuilder.customOperations") || [];
            return [customOperations.anyOf(this.component), customOperations.noneOf(this.component)].concat(filterBuilderCustomOperations);
        },

        publicMethods: function publicMethods() {
            return ["getCustomFilterOperations"];
        }
    };
}());

var DataControllerFilterSyncExtender = {
    skipCalculateColumnFilters: function skipCalculateColumnFilters() {
        var filterSyncEnabledValue = this.option("filterSyncEnabled");
        return filterSyncEnabledValue === "auto" ? this.option("filterPanel.visible") : filterSyncEnabledValue;
    },

    _calculateAdditionalFilter: function _calculateAdditionalFilter() {
        var that = this;

        if (that.option("filterPanel.filterEnabled") === false) {
            return that.callBase();
        }

        var filters = [that.callBase()],
            columns = that.getController("columns").getColumns(),
            filterValue = that.option("filterValue");

        if (that.skipCalculateColumnFilters()) {
            var currentColumn = that.getController("headerFilter").getCurrentColumn();
            if (currentColumn && filterValue) {
                filterValue = utils.syncFilters(filterValue, [currentColumn.dataField, "=", null], HEADER_FILTER_OPERATIONS);
            }
        }

        var calculatedFilterValue = utils.getFilterExpression(filterValue, columns, [customOperations.anyOf(), customOperations.noneOf()], "filterBuilder");
        if (calculatedFilterValue) {
            filters.push(calculatedFilterValue);
        }

        return gridCoreUtils.combineFilters(filters);
    },

    _parseColumnInfo: function _parseColumnInfo(fullName) {
        var matched = fullName.match(/columns\[([0-9]*)\]\.(.*)/);
        return {
            index: matched[1],
            changedField: matched[2]
        };
    },

    optionChanged: function optionChanged(args) {
        switch (args.name) {
            case "filterValue":
                this._applyFilter();
                var filterSyncController = this.getController("filterSync");
                if (!filterSyncController._skipSyncFilterValue && this.skipCalculateColumnFilters()) {
                    filterSyncController.syncFilterValue();
                }
                args.handled = true;
                break;
            case "filterSyncEnabled":
                args.handled = true;
                break;
            case "columns":
                if (this.skipCalculateColumnFilters()) {
                    var columnInfo = this._parseColumnInfo(args.fullName),
                        column;
                    if (["filterValues", "filterType"].indexOf(columnInfo.changedField) !== -1) {
                        column = this.getController("columns").getColumns()[columnInfo.index];
                        this.getController("filterSync").syncHeaderFilter(column);
                    } else if (["filterValue", "selectedFilterOperation"].indexOf(columnInfo.changedField) !== -1) {
                        column = this.getController("columns").getColumns()[columnInfo.index];
                        this.getController("filterSync").syncFilterRow(column, column.filterValue);
                    }
                }
                break;
            default:
                this.callBase(args);
        }
    }
};

var ColumnHeadersViewFilterSyncExtender = {
    _isHeaderFilterEmpty: function _isHeaderFilterEmpty(column) {
        if (this.getController("data").skipCalculateColumnFilters()) {
            return !utils.filterHasField(this.option("filterValue"), column.dataField);
        }

        return this.callBase(column);
    },

    _needUpdateFilterIndicators: function _needUpdateFilterIndicators() {
        return !this.getController("data").skipCalculateColumnFilters();
    },

    optionChanged: function optionChanged(args) {
        if (args.name === "filterValue") {
            this._updateHeaderFilterIndicators();
        } else {
            this.callBase(args);
        }
    }
};

module.exports = {
    defaultOptions: function defaultOptions() {
        return {
            /**
             * @name GridBaseOptions_filterValue
             * @publicName filterValue
             * @type Filter expression
             * @default null
             */
            filterValue: null,

            /**
             * @name GridBaseOptions_filterSyncEnabled
             * @publicName filterSyncEnabled
             * @type string|boolean
             * @default "auto"
             * @acceptValues "auto" | true | false
             */
            filterSyncEnabled: "auto"
        };
    },
    controllers: {
        filterSync: FilterSyncController
    },
    extenders: {
        controllers: {
            data: DataControllerFilterSyncExtender
        },
        views: {
            columnHeadersView: ColumnHeadersViewFilterSyncExtender
        }
    }
};